//iDataFilter
//by Alain Bonardi - 2004-2012
//

#include "ext.h"			// you must include this - it contains the external object's link to available Max functions
#include "ext_obex.h"		// this is required for all objects using the newer style for writing objects.

typedef struct _iDataFilter {	// defines our object's internal variables for each instance in a patch
	t_object p_ob;			// object header - ALL objects MUST begin with this...
	
	long n;			  	//nombre de valeurs dans le tampon//
	long n_echantillon;	//nombre d'échantillons 
	int *a;  		  	//vecteur de longueur n contenant les n dernières valeurs reçues//
	int *cl;		 	//les mêmes n valeurs mais classées dans l'ordre croissant//
	int *place;			//la place de chacune des valeurs de a dans le tableau cl//
	short dejaCree;	  	//à 0 si les matrices n'ont pas été créées, à 1 si elles l'ont été//
	short entreeOuverte; //à 1 si l'entrée laisse passer les valeurs, à 0 sinon//
	
	//
	//L'objet a une entrée : 0->valeur qui arrive // 1->état précédent V1 // 2->état suivant V2//
	//
	long p_value0; 	  //entrée 0 = valeur//
	long p_avg;
	long p_max;
	long p_min;
	long p_med;
	int sortant;
	long sigma;
	long milieu;
	long p_valeurs;	//nombre de valeurs entrées (à concurrence de n)//
	//
	//L'objet a 5 sorties : 
	//0-> moyenne entière//
	//1-> maximum entier//
	//2-> minimum entier//
	//3-> médiane entière//
	//4-> liste des valeurs en mémoire//
	void* outputIntAverage;
	void* outputIntMinimum;
	void* outputIntMaximum;
	void* outputIntMediane;
	void* outputDataList;
	
} t_iDataFilter;


// these are prototypes for the methods that are defined below
void iDF_creationVecteur(t_iDataFilter *x);
void iDF_changeNombreValeurs(t_iDataFilter *x, long m);
void iDF_prendEnCompteValeur(t_iDataFilter *x, long arg1);
 
void iDataFilter_assist(t_iDataFilter *x, void *b, long m, long a, char *s);

void *iDataFilter_new(long n);

void iDF_free(t_iDataFilter *x);
void iDF_assist(t_iDataFilter *x, void *b, long m, long a, char *s);
void iDF_dumpval(t_iDataFilter *x);
void iDF_dumpsorted(t_iDataFilter *x);
void iDF_dumpplace(t_iDataFilter *x);
void iDF_reset(t_iDataFilter *x);
void iDF_valeur(t_iDataFilter *x, long arg1);
void iDF_stackdepth(t_iDataFilter *x, long arg1);
void iDF_open(t_iDataFilter *x);
void iDF_close(t_iDataFilter *x);

t_class *iDataFilter_class;		// global pointer to the object class - so max can reference the object 


//--------------------------------------------------------------------------

int main(void)
{
	t_class *c;
	
	c = class_new("iDataFilter", (method)iDataFilter_new, (method)NULL, sizeof(t_iDataFilter), 0L, A_DEFLONG, 0);
	
	//liste des messages auquel on répond//
	
	//Arrivée de valeurs directement sur les entrées//
	class_addmethod(c, (method)iDF_valeur, "int",		A_LONG, 0);
	
	//message dumpval : vide la liste des valeurs sur la sortie liste//
	class_addmethod(c, (method) iDF_dumpval, "dumpval", A_NOTHING);
	
	//message dumpsorted : vide la liste des valeurs classées sur la sortie liste//
	class_addmethod(c, (method) iDF_dumpsorted, "dumpsorted", A_NOTHING);
	
	//message dumpplace : vide la liste des places des valeurs sur la sortie liste//
	class_addmethod(c, (method) iDF_dumpplace, "dumpplace", A_NOTHING);
	
	//message reset : réinitialise tout//
	class_addmethod(c, (method) iDF_reset, "reset", A_NOTHING);
	
	//message open : ouvre le canal//
	class_addmethod(c, (method) iDF_open, "open", A_NOTHING);
	
	//message close : ferme le canal//
	class_addmethod(c, (method) iDF_close, "close", A_NOTHING);
	
	//message stackdepth : reconfigure la taille de la pile//
	class_addmethod(c, (method) iDF_stackdepth, "stackdepth", A_LONG, 0);
	
	//on ajoute l'objet à la liste des objets Max//
	//finder_addclass("bonardiLib", "iDataFilter");
	// class_new() loads our external's class into Max's memory so it can be used in a patch
	// iDataFilter_new = object creation method defined below
	
	class_addmethod(c, (method)iDataFilter_assist,	"assist",	A_CANT, 0);	// (optional) assistance method needs to be declared like this
	
	 
	class_register(CLASS_BOX, c);
	iDataFilter_class = c;

	post("iDataFilter object loaded...",0);	// post any important info to the max window when our class is loaded
	return 0;
}


//--------------------------------------------------------------------------

void *iDataFilter_new(long arg1)		// n = int argument typed into object box (A_DEFLONG) -- defaults to 0 if no args are typed
{
	
	t_iDataFilter *x;				// local variable (pointer to a t_iDataFilter data structure)

	x = (t_iDataFilter *)object_alloc(iDataFilter_class); // create a new instance of this object
	
	/*
	intin(x,1);					// create a second int inlet (leftmost inlet is automatic - all objects have one inlet by default)
	x->p_outlet = intout(x);	// create an int outlet and assign it to our outlet variable in the instance's data structure
	
	x->p_value0	= 0;			// set initial (default) left operand value in the instance's data structure
	x->p_value1	= n;			// set initial (default) right operand value (n = variable passed to iDataFilter_new)
	*/
	
	//ATTENTION//
	//Il faut mettre les outlets dans l'ordre inverse de ce que l'on veut//
	x->outputDataList = listout(x);		//sortie list -> valeurs en mémoire//
	x->outputIntMediane = intout(x); 		//sortie int -> médiane//
	x->outputIntMaximum = intout(x);		//sortie int -> maximum//
	x->outputIntMinimum = intout(x); 		//sortie int -> minimum//
	x->outputIntAverage = intout(x); 		//sortie int -> médiane//
	
	if (arg1 > 0)
	{
		iDF_changeNombreValeurs(x, arg1);
	}
	else
	{
		iDF_changeNombreValeurs(x, 10);
	}
	iDF_reset(x);
	
	post(" new iDataFilter object instance added to patch...",0); // post important info to the max window when new instance is created
	
	return(x);					// return a reference to the object instance 
}


//--------------------------------------------------------------------------

void iDataFilter_assist(t_iDataFilter *x, void *b, long m, long a, char *s) // 4 final arguments are always the same for the assistance method
{
	/*
	if (m == ASSIST_OUTLET)
		sprintf(s,"Sum of Left and Right Inlets");
	else {
		switch (a) {	
		case 0:
			sprintf(s,"Inlet %ld: Left Operand (Causes Output)", a);
			break;
		case 1:
			sprintf(s,"Inlet %ld: Right Operand (Added to Left)", a);
			break;
		}
	}
	 */
}



//*********************************************************************************//
//void creationVecteur(t_iDataFilter *x)
//
//Crée un vecteur a de la longueur n, ainsi que les vecteurs de classement et de   //
//place des valeurs arrivées en entrée											   //		  
//*********************************************************************************//	
void iDF_creationVecteur(t_iDataFilter *x)
{
	x->a = (int *) calloc(x->n, sizeof(int)); //création de l'espace mémoire nécessaire//
	x->cl = (int *) calloc(x->n, sizeof(int));
	x->place = (int *) calloc(x->n, sizeof(int));
	x->p_min = INT_MAX;
	x->p_valeurs = 0;
  	if (x->a == NULL)
	{
   		error("problem for creating the sample vector of size %d", x->n);
	}
  	else
	{
   		post("Creation of the sample vector of size %d is OK", x->n);
	}
   	if ((x->a != NULL) && (x->cl != NULL) && (x->place != NULL))
	{
		iDF_reset(x);
	}
}

//*********************************************************************************//
//void changeNombreValeurs(t_iDataFilter *x, long m)
//
//Change le nombre de valeurs en entrée, et reconfigure en conséquence les vecteurs//
//*********************************************************************************//
void iDF_changeNombreValeurs(t_iDataFilter *x, long m)
{
	x->n = m;
	if (x->dejaCree == 1)
	{
		free(x->a);
		free(x->cl);
		free(x->place);
		x->dejaCree = 0;
	};
	//création des deux matrices//
	iDF_creationVecteur(x);
	x->dejaCree = 1;
}

//*********************************************************************************//
//void prendEnCompteValeur(t_iDataFilter *x, long arg1)
//
//Prend en compte la valeur en entrée, en procédant à un décalage dans a		   //
//à la mise à jour du tableau de classement et des places						   //
//Sort la moyenne, le minimum, le maximum et la valeur médiane sur les sorties	   //			  
//*********************************************************************************//
void iDF_prendEnCompteValeur(t_iDataFilter *x, long arg1) {
	int i, p, np, r;
	//
	//on conserve la dernière valeur//
	x->sortant = *(x->a+x->n-1);
	if (x->p_valeurs < x->n)
		x->p_valeurs += 1;
	p = *(x->place+x->n-1); //on garde la place dans le tableau classé du sortant//
	//post("place du sortant = %d", p);
	//post("cl avant toute opération");
	//t_iDataFilter_dumpsorted(x);
	//on supprime le sortant dans le tableau cl//
	//on modifie les places des éléments en conséquence//
	for (i = p; i < x->n-1; i++)
	{
		*(x->cl+i) = *(x->cl+i+1);
	}
	//provisoirement, on met 0 dans la dernière case du classement//
	*(x->cl+x->n-1) = 0;
	//post("après suppression du sortant et décalage vers la dans cl");
	//t_iDataFilter_dumpsorted(x);
	//on modifie le tableau de placement//
	//en balayant les valeurs, et en regardant celles qui étaient plus grandes que p//
	//que l'on diminue de 1//
	for (i = 0; i < x->n; i++)
	{
		if (*(x->place+i) > p)
		{
			(*(x->place+i))--;
		}
	}
	//on décale toutes les valeurs d'un cran vers la droite dans les tableaux a et place//
	for (i = x->n-1; i >= 0; i--)
	{
		*(x->a+i) = *(x->a+i-1);
		*(x->place+i) = *(x->place+i-1);
	}
	//on insère la nouvelle valeur en tête de tableau
	*(x->a) = arg1;
	//on cherche la place (np) de la nouvelle valeur arg1 dans le tableau cl//
	np = 0;
	while ((*(x->cl+np) < arg1) && (np < x->n-1))
	{
		np++;
	}
	//post("place np = %d", np);
	//le nouvel entrant est inséré à sa place//
	//on fait d'abord de la place en décalant les éléments vers la droite//
	for (i = x->n-1; i > np; i--)
	{
		*(x->cl+i) = *(x->cl+i-1);
	}
	//post("après avoir décalé vers la droite pour faire place au nouveau");
	//t_iDataFilter_dumpsorted(x);
	//post("avant mise à jour des places suite au décalage");
	//t_iDataFilter_dumpplace(x);
	for (i = 0; i < x->n; i++)
	{
		if (*(x->place+i) >= np)
		{
			(*(x->place+i))++;
		}
	}
	//la place du nouvel entrant est mémorisée//
	*(x->place) = np;
	//le nouvel entrant est casé//
	*(x->cl+np) = arg1;
	//post("après avoir inséré la nouvelle valeur");
	//t_iDataFilter_dumpsorted(x);
	//post("nouvelles places");
	//t_iDataFilter_dumpplace(x);
	//on met à jour les rubriques en sortie//
	//la somme (sigma) des valeurs est rentrée du sortant et augmentée de l'entrant//
	x->sigma -= x->sortant;
	x->sigma += arg1;
	x->p_avg = x->sigma / x->p_valeurs;
	r = x->sigma%x->n;
	if (r >= (x->n)/2)
	{
		x->p_avg++;
	}
	outlet_int(x->outputIntAverage, x->p_avg);
	//
	x->p_max = *(x->cl+x->n-1);
	outlet_int(x->outputIntMaximum, x->p_max);
	//
	//x->p_min = *(x->cl+0);
	if (arg1 < x->p_min)
		x->p_min = arg1;
	outlet_int(x->outputIntMinimum, x->p_min);
	//
	x->p_med = *(x->cl+(x->milieu));
	outlet_int(x->outputIntMediane, x->p_med);
}

//*********************************************************************************//
//void t_iDataFilter_stackdepth(t_iDataFilter *x, long arg1)
//
//
//*********************************************************************************//
void iDF_stackdepth(t_iDataFilter *x, long arg1) {
	iDF_changeNombreValeurs(x, arg1);
}

//*********************************************************************************//
//void iDF_reset(t_iDataFilter *x)
//
//Remet tout à 0 et réinitialise les variables									   //	
//*********************************************************************************//
void iDF_reset(t_iDataFilter *x) {
	int i;
	//
	for (i = 0; i < x->n; i++)
	{
   		*(x->a+i) = 0;
   		*(x->cl+i) = 0;
   		*(x->place+i) = i;
	}
   	x->sigma = 0;
   	x->p_value0 = 0;
	x->dejaCree = 0;
	x->entreeOuverte = 1;
	x->p_min = INT_MAX;
	x->p_valeurs = 0;
	if ((x->n)%2 == 0)
	{
		//si le nombre de valeurs est pair, le milieu est en n/2-1 (démarre en 0)
		x->milieu = (x->n)/2-1;
	}
	else
	{
		//si le nombre de valeurs est impair, le milieu est en n/2 (démarre en 0)
		x->milieu = (x->n)/2;
	}
}

//*********************************************************************************//
//void iDF_open(t_iDataFilter *x)
//
//Ouvre l'entrée de prise en compte des valeurs									   //	
//*********************************************************************************//
void iDF_open(t_iDataFilter *x) {
	x->entreeOuverte = 1;
}

//*********************************************************************************//
//void iDF_close(t_iDataFilter *x)
//
//Ferme l'entrée de prise en compte des valeurs									   //	
//*********************************************************************************//
void iDF_close(t_iDataFilter *x) {
	x->entreeOuverte = 0;
}


//*********************************************************************************//
//void iDF_valeur(t_iDataFilter *x, long arg1)
//
//Prise en compte d'une valeur entière sur l'entrée								   //	
//*********************************************************************************//
void iDF_valeur(t_iDataFilter *x, long arg1) {
	if (x->entreeOuverte == 1)
	{
		x->p_value0 = arg1;
		iDF_prendEnCompteValeur(x, arg1);
	}
}

//*********************************************************************************//
//void iDF_free(t_iDataFilter *x)
//
//Libère les ressources avant de supprimer l'objet								   //
//*********************************************************************************//
void iDF_free(t_iDataFilter *x)
{
	if (x->dejaCree == 1)
	{
		free(x->a);
		free(x->cl);
		free(x->place);
		x->dejaCree = 0;
		//post("Sample vector deleted");
	};
}

//*********************************************************************************//
//void iDF_dumpval(t_iDataFilter *x)
//
//Vide le contenu de la table des valeurs										   //
//*********************************************************************************//
void iDF_dumpval(t_iDataFilter *x) {
	t_atom *myList;
	short i;
	//
	myList = (t_atom *) calloc(x->n, sizeof(t_atom)); //création de l'espace mémoire nécessaire//
	
	if (myList != NULL)
	{
		for (i=0; i<x->n; i++)
		{
			SETLONG(myList+i, *(x->a+i));	
		}
		outlet_list(x->outputDataList, 0L, x->n, myList);
		free(myList);
	}
}


//*********************************************************************************//
//void iDF_dumpsorted(t_iDataFilter *x)
//
//Vide le contenu de la table triée												   //	
//*********************************************************************************//
void iDF_dumpsorted(t_iDataFilter *x) {
	t_atom *myList;
	short i;
	//
	myList = (t_atom *) calloc(x->n, sizeof(t_atom)); //création de l'espace mémoire nécessaire//
	
	if (myList != NULL)
	{
		for (i=0; i<x->n; i++)
		{
			SETLONG(myList+i, *(x->cl+i));	
		}
		outlet_list(x->outputDataList, 0L, x->n, myList);
		free(myList);
	}
}



//*********************************************************************************//
//void iDF_dumpplace(t_iDataFilter *x)
//
//Vide le contenu de la table des places des valeurs entrées dans la table triée   //	
//*********************************************************************************//
void iDF_dumpplace(t_iDataFilter *x) {
	t_atom *myList;
	short i;
	//
	myList = (t_atom *) calloc(x->n, sizeof(t_atom)); //création de l'espace mémoire nécessaire//
	
	if (myList != NULL)
	{
		for (i=0; i<x->n; i++)
		{
			SETLONG(myList+i, *(x->place+i));	
		}
		outlet_list(x->outputDataList, 0L, x->n, myList);
		free(myList);
	}
}

